# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/90_lens.ipynb.

# %% auto 0
__all__ = ['trylen', 'aopn', 'lenn', 'hasn', 'has1', 'has2', 'has3', 'has4']

# %% ../nbs/90_lens.ipynb 4
from operator import ge
from functools import wraps, partial

from types import BuiltinFunctionType
from typing import TypeGuard
from typing import Iterable, Callable, Optional, Union

# %% ../nbs/90_lens.ipynb 5
from chck import isint, ismthdwrap

# %% ../nbs/90_lens.ipynb 6
from .atyp import Operator
from .core import make

# %% ../nbs/90_lens.ipynb 10
def trylen(a: int | Iterable, n: int = 1, op: Operator = ge) -> TypeGuard[Iterable]:
    '''Check if `op(a, n)` if a is an `int` otherwise tries `op(len(a), n)` otherwise `False`.'''
    if ismthdwrap(a): a = a()
    if isint(a): return op(a, n)
    try: return op(len(a), n)
    except: return False

# %% ../nbs/90_lens.ipynb 12
def aopn(n: int, op: Operator = ge, __fn: str = 'len') -> Callable:
    f'''Factory for the function  `def func(a): return op(len(a), {n})`.'''
    @wraps(trylen, assigned=('__doc__'))
    def wrap(a: int | Iterable) -> TypeGuard[Iterable]:
        f'''Check if `{op.__name__}(len(a), {n}) == True.'''
        return trylen(a, n, op)
    for a in ('__name__', '__qualname__'): setattr(wrap, a, f'{__fn}{op.__name__}{n}')
    return wrap

# %% ../nbs/90_lens.ipynb 15
def lenn(n: int, op: Optional[Union[BuiltinFunctionType, Callable]] = ge) -> Callable:    
    func = partial(aopn, n, op or ge, __fn='len')()
    @wraps(func, assigned=('__doc__', '__annotations__'))
    def wrap(a: int | Iterable) -> TypeGuard[Iterable]:
        f'''Check if `a` contain at least `{n}` elements.'''
        return func(a)
    return wrap

# %% ../nbs/90_lens.ipynb 18
def hasn(a, n: int = 1) -> TypeGuard[Iterable]:
    '''Check if `a` contain at least `n` elements.'''
    func = partial(lenn, n, ge, __fn='has')()
    @wraps(func, assigned=('__doc__', '__annotations__'))
    def wrap(a: int | Iterable) -> TypeGuard[Iterable]:
        f'''Check if `a` contain at least `{n}` elements.'''
        return func(a)
    return wrap(a)

# %% ../nbs/90_lens.ipynb 20
@make('__len__', guard = lenn(1))
def has1(a) -> TypeGuard[Iterable]: '''Check if `a` has at least `1` elements.'''

@make('__len__', guard = lenn(2))
def has2(a) -> TypeGuard[Iterable]: '''Check if `a` has at least `2` elements.'''

@make('__len__', guard = lenn(3))
def has3(a) -> TypeGuard[Iterable]: '''Check if `a` has at least `3` elements.'''

@make('__len__', guard = lenn(4))
def has4(a) -> TypeGuard[Iterable]: '''Check if `a` has at least `4` elements.'''



__version__ = "0.0.1"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00__init__.ipynb 3
from itertools import chain
from . import (atyp, core, attr, call, lens,)

from .atyp import *
from .core import *
from .attr import *
from .call import *
from .lens import *

# %% ../nbs/00__init__.ipynb 4
__all__ = sorted(set(chain(*[
    mod.__all__ for mod in (atyp, core, attr, call, lens,)
])))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_attr.ipynb.

# %% auto 0
__all__ = ['hasshape', 'hascolumns', 'hasdf', 'hasindex']

# %% ../nbs/10_attr.ipynb 4
from typing import TypeGuard
from typing import Callable

# %% ../nbs/10_attr.ipynb 5
from .core import make

# %% ../nbs/10_attr.ipynb 7
@make('shape')
def hasshape(obj) -> TypeGuard[Callable]: '''Checks if `obj.shape` exists'''

@make('columns')
def hascolumns(obj) -> TypeGuard[Callable]: '''Checks if `obj.columns` exists'''

@make('df')
def hasdf(obj) -> TypeGuard[Callable]: '''Checks if `obj.df` exists'''

@make('index')
def hasindex(obj) -> TypeGuard[Callable]: '''Checks if `obj.index` exists'''



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/hasr',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/hasr',
                'lib_path': 'hasr'},
  'syms': { 'hasr.attr': { 'hasr.attr.hascolumns': ('attr.html#hascolumns', 'hasr/attr.py'),
                           'hasr.attr.hasdf': ('attr.html#hasdf', 'hasr/attr.py'),
                           'hasr.attr.hasindex': ('attr.html#hasindex', 'hasr/attr.py'),
                           'hasr.attr.hasshape': ('attr.html#hasshape', 'hasr/attr.py')},
            'hasr.atyp': {},
            'hasr.call': { 'hasr.call.hasget': ('call.html#hasget', 'hasr/call.py'),
                           'hasr.call.haskeys': ('call.html#haskeys', 'hasr/call.py')},
            'hasr.core': { 'hasr.core.hasr': ('core.html#hasr', 'hasr/core.py'),
                           'hasr.core.make': ('core.html#make', 'hasr/core.py'),
                           'hasr.core.name': ('core.html#name', 'hasr/core.py'),
                           'hasr.core.qual': ('core.html#qual', 'hasr/core.py')},
            'hasr.lens': { 'hasr.lens.aopn': ('lens.html#aopn', 'hasr/lens.py'),
                           'hasr.lens.has1': ('lens.html#has1', 'hasr/lens.py'),
                           'hasr.lens.has2': ('lens.html#has2', 'hasr/lens.py'),
                           'hasr.lens.has3': ('lens.html#has3', 'hasr/lens.py'),
                           'hasr.lens.has4': ('lens.html#has4', 'hasr/lens.py'),
                           'hasr.lens.hasn': ('lens.html#hasn', 'hasr/lens.py'),
                           'hasr.lens.lenn': ('lens.html#lenn', 'hasr/lens.py'),
                           'hasr.lens.trylen': ('lens.html#trylen', 'hasr/lens.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/20_call.ipynb.

# %% auto 0
__all__ = ['hasget', 'haskeys']

# %% ../nbs/20_call.ipynb 4
from typing import TypeGuard
from typing import Callable

# %% ../nbs/20_call.ipynb 5
from .core import make

# %% ../nbs/20_call.ipynb 7
@make('get', guard=callable)
def hasget(obj) -> TypeGuard[Callable]:
    '''Checks if `obj.get` is a callable'''


@make('keys', guard=callable)
def haskeys(obj) -> TypeGuard[Callable]: 
    '''Checks if `obj.keys` is a callable'''



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_core.ipynb.

# %% auto 0
__all__ = ['hasr', 'qual', 'name', 'make']

# %% ../nbs/01_core.ipynb 4
from functools import wraps, update_wrapper
from typing import TypeGuard
from typing import Any, Callable, Optional, ParamSpec, TypeVar, overload, get_type_hints, get_overloads

# %% ../nbs/01_core.ipynb 5
from chck import isnone, notnone

# %% ../nbs/01_core.ipynb 7
from .atyp import P, R

# %% ../nbs/01_core.ipynb 8
def hasr(
    obj: object, attr: str, check: bool = True, 
    guard: Optional[Callable] = notnone, default = None
) -> TypeGuard[object]:
    '''Check if an object has a specified attribute with an optional value check.

    This function tests if an object has an attribute with the option to validate 
    the attribute's value using a guard function. The guard function can be any callable 
    that takes a single argument and returns a boolean. If the attribute is not present, 
    a default value can be considered for the guard check.

    Parameters
    ----------
    obj : object
        The object to check for the attribute.

    attr : str
        The name of the attribute to check for.

    check : bool, optional
        If True, the function performs a guard check on the attribute's value, by default True.

    guard : Optional[Callable], default: `notnone`.
        The guard function used to validate the attribute's value, by default notnone.

    default : Any, optional
        The default value to use if the attribute is not found, by default None.

    Returns
    -------
    TypeGuard[object]
        Returns True if the object has the attribute (and passes the guard check if `check` is True),
        otherwise False.

    Examples
    --------
    >>> d = dict(a=1, b=2, c=3)
    >>> hasr(d, 'keys')
    True

    >>> hasr(d, 'missing', default=10)
    False

    See Also
    --------
    chck.notnone : A guard function that checks if a value is not None i.e. `val is not None`.
    '''
    has = hasattr(obj, attr)
    got = getattr(obj, attr, default)
    if check and callable(guard):
        try: return has and guard(got)
        except: False
    return has and notnone(got)

# %% ../nbs/01_core.ipynb 9
def qual(attr: str, base: str = 'has', prefix: str = '', suffix: str = '') -> str:
    '''Produce a '__qualname__' for a function based on an attribute name.'''
    if isnone(prefix): prefix = ''
    if isnone(suffix): suffix = ''
    return f'{prefix}{base}{attr}{suffix}'
    
def name(fn: Callable, attr: str, base: str = 'has', prefix: str = None, suffix: str = None) -> Callable:
    for a in ('__name__', '__qualname__'): setattr(fn, a, qual(attr, base, prefix, suffix))
    return fn

# %% ../nbs/01_core.ipynb 10
def make(
    attr: str, check: bool = True, 
    guard: Optional[Callable] = notnone, 
    default = None
) -> Callable[[R], R]:
    '''Create a decorator for generating attribute-checking functions.

    This decorator factory returns a decorator that, when applied to a function,
    creates a new function which checks if an object has a specified attribute (named `attr`).
    Additionally, it can validate the attribute's value using a guard function `guard`.
    The decorator updates the original function's docstring to include the new parameters.

    Parameters
    ----------
    attr : str
        The name of the attribute to check for in the decorated function.

    check : bool, default: True
        If True, the decorated function performs a guard check on the attribute's value.

    guard : Optional[Callable], default: notnone
        The guard function used to validate the attribute's value in the decorated function. 
        Defaults to `notnone`, which checks that the value is not None.

    default : Any, optional
        The default value to use if the attribute is not found in the decorated function.

    Returns
    -------
    Callable[[R], R]
        A decorator that, when applied to a function, creates a new attribute-checking function.
        This new function includes additional parameters and updated docstring.

    Examples
    --------
    >>> @make('keys', guard=callable)
    ... def haskeys(obj) -> TypeGuard[Callable]: 
    ...     """Checks if `obj.keys` is a callable"""

    >>> d = dict(a=1)
    >>> haskeys(d), haskeys([])
    (True, False)
    '''
    def deco(fn: Callable[..., TypeGuard[object]]):
        fndoc = getattr(fn, '__doc__', None)
        if isnone(fndoc): fndoc = f'Checks if an object has {attr!r} i.e. `obj.{attr}`.'
        doc = f'''
            {fndoc}

            Parameters
            ----------
            obj : object
                The object to check for {attr!r} i.e. `obj.{attr}`.
                
            check : bool, default: {check}
                If True, the function performs a typeguard check on the `obj.{attr}` value.
        
            guard : Optional[Callable], default: `notnone`.
                The guard function used to validate `obj.{attr}`'s value if `check == True`.
        
            default : Any, optional
                The default value to use if `obj` does not have {attr!r}..
        
            Returns
            -------
            TypeGuard[object]
                Returns True if the object has {attr!r} (i.e. `obj.{attr}` and passes the guard check if `check` is True),
                otherwise False.
        
            Examples
            --------
            >>> d = dict(a=1, b=2, c=3)
            >>> hasr(d, 'keys')
            True
        
            >>> hasr(d, 'missing', default=10)
            False
        
            See Also
            --------
            chck.notnone : A guard function that checks if a value is not None i.e. `val is not None`.
            '''
        @wraps(fn, assigned=('__module__', '__name__', '__qualname__', ))
        def wrap(
            obj: object, check: bool = check, 
            guard: Optional[Callable] = guard, 
            default = default
        ) -> TypeGuard[object]:
            return hasr(obj, attr, check=check, guard=guard, default=default)
        wrap.__doc__ = doc
        wrap = name(wrap, attr)        
        return wrap
    return deco



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_atyp.ipynb.

# %% auto 0
__all__ = ['P', 'R', 'Operator']

# %% ../nbs/00_atyp.ipynb 4
from types import BuiltinFunctionType
from typing import Callable, Union, TypeAlias, ParamSpec, TypeVar, TypeGuard

# %% ../nbs/00_atyp.ipynb 6
P = ParamSpec('P')
R = TypeVar('R', bound=Callable[..., TypeGuard[object]])

# %% ../nbs/00_atyp.ipynb 7
Operator: TypeAlias = Union[BuiltinFunctionType, Callable]
